//
// Eradial.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task to extract a radial profile from an image
//              and compare it to the theoretical PSF
//
//#include <ctime>

#include "CalServer.h"
#include "Dal.h"

#include "Eradial.h"
#include "ImageRegion.h"
#include "CalSetter.h"
#include "STime.h"
#include "UpperLimitSolver.h"

#include "Tracer.h"
#include "errstr.h"
#include "param.h"
#include "slatec.h"
//gcc4.8.2 portability
#include <unistd.h>


/* Method: Eradial() - Constructor */
Eradial::Eradial():
srcreg(0),
_calset(0)
{
  PROC;

  /* Get an extension for temporary files based on time and pid */
    //srand(time(NULL) + getpid());    // sets the seed for the 
     srand(getpid());                  // random numbers by using the 
                                     // system time and the process id
    //cout<<"Random: "<<getpid()<<" "<<time(NULL)<<endl;

    _postfix = toString(rand());

    // _postfix = toString(STime().nowInMillisec());

  /* Get the parameters from this command line */
    string imagefile = stringParameter("imageset");
    string srcexp = stringParameter("srcexp");
    double rpbinwid = realParameter("binwidth"); // Width of rad prof bins "
    double energy = realParameter("psfenergy") * 1000.; // Width of rad prof bins "
    string psftype = stringParameter("psftype");
    bool centroid = booleanParameter("centroid"); // Is centroiding wanted ?
    double maxrad = realParameter("maxradius"); // maximum radius in arcsecs

  /* Attempt to set the CAL state using the input image header */
    _calset = new CalSetter();
    _calset->fromImage(imagefile);
    setPsfType(psftype);

  /* We can't currently get radial profiles from the Fast modes so
     set status bad and exit early in these cases */
    Xmm::Mode windowMode = calServer.state()->mode()->get();

    if (windowMode == Xmm::FAST_TIMING ||
        windowMode == Xmm::FAST_BURST ||
        windowMode == Xmm::FAST_UNCOMPRESSED || 
        windowMode == Xmm::FAST_COMPRESSED) {
       
       setParameter("status",false);

       errstr<<" Timing/Burst mode is not currently supported"
          << Sas::message(Msg::App, Msg::Verbose);

       ALTRETURN;
    }

  /* Set up an ImageRegion object for the source area */
    errstr<<" Setting up source image region"
          << Sas::message(Msg::App, Msg::Verbose);

    srcreg = new ImageRegion(imagefile, srcexp, 
                                "filtpixlist_src" + _postfix + ".ds:");

    errstr<<" Set up source image region"
          << Sas::message(Msg::App, Msg::Verbose);

  /* Check out a few methods */
  /*
    double delrad = srcreg->getRadius();
    double deltheta = srcreg->theta();
    double delcnts = srcreg->counts();
    double delflux = srcreg->flux();

    errstr<<" Radius: "<<delrad
          <<" Theta: "<<deltheta
          <<" Phi: "<<srcreg->phi() << endl
          <<" Counts: "<<delcnts << endl
          <<" numpix: "<<srcreg->number_of_pixels() << endl
          <<" flux: "<<delflux << endl
          <<" pixelArea: "<<srcreg->pixelArea()
          <<" xwid: "<<srcreg->xWidth()
          <<" ywid: "<<srcreg->yWidth()
          << Sas::message(Msg::App, Msg::Verbose);
  */

  /* Get the source centre, centroiding if requested */
    CoordPair centre=srcreg->centre();
    if (centroid)
       centre = srcreg->centroid(5);

  /* Get the radial profile */
    ImageRegion::RadialProfile r = srcreg->getRadProf(centre.x(),
                                      centre.y(), rpbinwid, maxrad);


  /* Get the corresponding PSF function */
    vector<double> psfdata = calcPsfArray(r, energy, 
                                 srcreg->theta(), srcreg->phi());

  /* Fit the data using a linear least squares fit */
    int maxDegree = 2;
    double rms = 1.e-20;
    int ierr, degree;               //      error code from SLATEC
    int nPoints = r.data.size();
    double  *weights = new double [nPoints],
            *bestFitPSF = new double [nPoints],
            *xp = new double [nPoints],
            *yp = new double [nPoints],
            *linCoeffs = new double [3*nPoints+3*maxDegree+3];

    for (int n=0; n<nPoints; ++n) {
            yp[n] = r.data[n]; xp[n] = psfdata[n]; 

          /* Set weights to 1.0 / relative error */
            if (r.error[n] > 0.0) {
                 //weights[n] = 1.0 / (r.error[n] / r.data[n]);
                 weights[n] = 1.0 / (r.error[n]);
                 //weights[n] = r.data[n] / r.error[n];
                 //weights[n] = 0.0001;
            } else {
                 weights[n] = 0.0001;  // to say something
            }
            //cout<<xp[n]<<" "<<yp[n]<<" "<<weights[n]<<endl;

    }

//    vector<double> bestFitPSF(nPoints);
//    vector<double> linCoeffs(3*nPoints+3*maxDegree+3);
  //  SLATEC_DPOLFT(nPoints, r.data, psfdata, r.error, maxDegree, degree, rms,
  //                                            bestFitPSF, ierr, linCoeffs);
    SLATEC_DPOLFT(nPoints, xp, yp, weights, maxDegree, degree, rms,
                                              bestFitPSF, ierr, linCoeffs);

    errstr<<"Error status from DPOLFT: "<<ierr
          << Sas::message(Msg::App, Msg::Verbose);
    errstr<<"RMS fit statistic: "<<rms
          << Sas::message(Msg::App, Msg::Always);
    errstr<<"Background fit as: "<<linCoeffs[1]
          << Sas::message(Msg::App, Msg::Always);
  /*
LSFIT_result
    LSFIT_result result = linearfit(psfdata,r.data,r.error);

    errstr<<"PSF fit statistic: "<<result[0]<<endl
          <<"Normalisation: "<<result[1]<<endl
          <<"Background: "<<result[2]<<endl
          <<Sas::message(Msg::App,Msg::Always);
  */
   /*
    for (int i=0; i<r.nbins; i++)
       
       cout<<(i+0.5)*r.rwid<<" "<<r.data[i]<<" "<<r.error[i]
           <<" "<<psfdata[i]<<" "<<(r.data[i]-linCoeffs[1])
           <<" "<<psfdata[i]*linCoeffs[2]<<endl;
   */

  /* Output the results to a FITS file */
    string rpsetname = stringParameter("radprofset");
    DataSet *  rpset = dataSetServer->open(rpsetname, DataSet::Create);

    writerpfile(rpset, r, psfdata, bestFitPSF, linCoeffs, rms,
                srcreg->attributes());

  // Clean up
    errstr
        << "Closing Radial Profile set " << rpsetname 
        << Sas::message(Msg::App, Msg::Sparse);
    dataSetServer->close(rpset);

  ENDPROC;
}

/* Destructor */
Eradial:: ~Eradial()
{
  PROC;

    delete srcreg;            srcreg=0;
    delete _calset;           _calset=0;

  ENDPROC;
};

/* Method: getRadialProfile - calculates the radial profile of the source
                              and the PSF */
void Eradial::setPsfType(string & psfType) const
{
  PROC;

  /* Set up the encircled energy function */
   if (psfType == "LOW")
      _calset->setAccuracyLevel(LOW);
   else if (psfType == "MEDIUM")
      _calset->setAccuracyLevel(MEDIUM);
   else if (psfType == "HIGH")
      _calset->setAccuracyLevel(HIGH);
   else if (psfType == "EXTENDED")
      _calset->setAccuracyLevel(EXTENDED);
   else if (psfType == "ELLBETA")
      _calset->setAccuracyLevel(ELLBETA);

  ENDPROC;
}

/* Method: calcPsfArray - calculates the PSF profile in the bins
                          given in the radial profile object */
vector<double> Eradial::calcPsfArray(ImageRegion::RadialProfile r, 
                            double energy, double theta, double phi) 
{
 PROC;

   PsfDataServer * psfs = 0; psfs = calServer.getAtom(psfs);
   Psf & p = psfs->psf(energy, theta, phi);
   Psf * _psf = &p;

   vector<double> psfsb(r.nbins);

  /* Set the conversion from arcsecs to mm */
   //double asec2mm = _psf->px();
   double asec2mm = 27.5;
   double binwidmm = r.rwid / asec2mm;

  /* Loop over each radial profile bin and get the PSF surface brightness.
     Don't worry about normalisation, this will be done later during
     the fitting */
   for (int i=r.nbins-1; i>=0; i--)   
   {
    /* Find the outer and inner radii in mm */
       double inner = (i) * binwidmm;
       double outer = (i+1) * binwidmm;
       double ann_area = (outer * outer - inner * inner);

    /* Get the encircled energies for these radii */
       double eeinner = _psf->encircledEnergy(inner);  // could save time here
       double eeouter = _psf->encircledEnergy(outer);

    /* Find the surface brightness within this annulus */
       psfsb[i] = (eeouter - eeinner) / ann_area;
   }

   delete _psf; _psf = 0;

   RETURN psfsb;
}

const void  Eradial::writerpfile(DataSet * rpset, ImageRegion::RadialProfile r,
          vector<double> psfdata, double * bestFitPSF, double * linCoeffs,
          double rms, DatafileAttributes * imgatts) const
{
     // Add the data table to the dataset
      Table * datatable = rpset->addTable("RPSF", r.nbins);

     // Add the columns relevant for an OGIP-compliant radial profile
      Column * rlowCol = datatable->addColumn("RAD_LO", Column::Real32,
                                  "start of radial bin", "arcsec");
      Column * rhiCol = datatable->addColumn("RAD_HI", Column::Real32,
                                  "end of radial bin", "arcsec");
      Column * rpCol = datatable->addColumn("RPROF", Column::Real32,
                                  "Radial Profile", "count/arcsec**2");
      Column * rperrCol = datatable->addColumn("RPROF_ERR", Column::Real32,
                                  "Radial Profile error", "count/arcsec**2");
      Column * psfCol = datatable->addColumn("RPSF", Column::Real32,
                                  "Point Spread Function", "count/arcsec**2");
      Column * bfpsfCol = datatable->addColumn("FIT_RPSF", Column::Real32,
                         "Best fit Point Spread Function", "count/arcsec**2");
      ColumnData * rlowColData   = rlowCol->data();
      real32     * rlow      = rlowColData->real32Data();
      ColumnData * rhiColData   = rhiCol->data();
      real32     * rhi      = rhiColData->real32Data();
      ColumnData * rpColData   = rpCol->data();
      real32     * rp      = rpColData->real32Data();
      ColumnData * rperrColData   = rperrCol->data();
      real32     * rperr      = rperrColData->real32Data();
      ColumnData * psfColData   = psfCol->data();
      real32     * psfout      = psfColData->real32Data();
      ColumnData * bfpsfColData   = bfpsfCol->data();
      real32     * bfpsfout      = bfpsfColData->real32Data();

     // Write the data
      for (int i=0; i<r.nbins; i++)
      {
          rlow[i] = i * r.rwid; 
          rhi[i] = (i+1) * r.rwid; 
          rp[i] = r.data[i];   // radial profile data
          rperr[i] = r.error[i]; // error on rp data
          psfout[i] = psfdata[i]; // The raw PSF
          bfpsfout[i] = bestFitPSF[i]; // The best fit PSF
      }

     /* Clean up the column pointers */
      delete rlowColData; delete rhiColData; delete rpColData;
      delete rperrColData; delete psfColData; delete bfpsfColData;

    // Add the attributes appropriate for an OGIP-compliant radial profile
      *datatable->addAttribute("TELESCOP",    "mission/satellite name") =
                               imgatts->telescope();
        
      *datatable->addAttribute("INSTRUME",    "instrument name") =
                               imgatts->instrument();

      *datatable->addAttribute("FILTER",      "filter in use")         = 
                               imgatts->filter();

      *datatable->addAttribute("SUBMODE",      "filter in use")         = 
                               imgatts->observingMode();

      *datatable->addAttribute("ARFVERSN",    "OGIP classification of FITS format")      = "1992a";

      *datatable->addAttribute("HDUCLASS",    "Format conforms to OGIP standard")        = "OGIP";

      *datatable->addAttribute("HDUCLAS1",    "dataset relates to radial response")    = "RESPONSE";

      *datatable->addAttribute("HDUCLAS2",    "dataset is a radial response")   = "RPRF";
      *datatable->addAttribute("HDUCLAS3",    "the radial profile contains bckgnd")   = "TOTAL";

      *datatable->addAttribute("HDUVERS1",    "Obsolete - included for backwards compatibility")     = "1.0.0";
      *datatable->addAttribute("HDUVERS2",    "Version of format")     = "1.0.1";
      *datatable->addAttribute("FIT_0", "Constant term of linear fit of PSF to rad prof.")     = linCoeffs[1];
      *datatable->addAttribute("FIT_1", "Linear term of linear fit of PSF to rad prof.")     = linCoeffs[2];
      *datatable->addAttribute("FIT_STAT", "RMS error in linear fit of PSF to rad prof.")     = rms;

}
      

