//
// Eupper.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task to calculate the optimum radius and centroid
//              for a point source extraction
//
//#include <ctime>
#include <iostream>
#include <fstream>
//gcc4.8.2 portability
#include <unistd.h>


#include "Tracer.h"
#include "errstr.h"
#include "param.h"

#include "Dal.h"
#include "Eupper.h"
#include "UpperLimitSolver.h"
#include "ArrayPixelIterator.h"
#include "SscWcsConvertor.h"

/* Method: Eupper() - Constructor */
Eupper::Eupper()
{
  PROC;

    //cout<<"CLOCK: Starting: "<<clock();
  /* Get the parameters from this command line */
    string imagefile = stringParameter("imageset");
    string exposurefile = stringParameter("exposuremap");
    double src_ra = realParameter("srcra");
    double src_dec = realParameter("srcdec");
    double src_rad = realParameter("srcradius");
    double bck_ra = realParameter("bgdra");
    double bck_dec = realParameter("bgddec");
    double bck_rad_in = realParameter("bgdrinner");
    double bck_rad_out = realParameter("bgdrouter");
    double ulsig = realParameter("ulsig");

  // for (unsigned n=0; n<100; ++n) {   // REMOVE

   // imagefile = "imagepnxy"+toString(n)+".ds";   // REMOVE
   // exposurefile = "imagepnxyexp"+toString(n)+".ds";   // REMOVE

  /* Get a pixel iterator for the image */
    ImageData srcImage = readImage(imagefile);
 
  /* Calculate the counts in the source region */
    COUNTS source = counts(srcImage, src_ra, src_dec, 0.0, src_rad);
    
  /* Calculate the counts in the background region */
    COUNTS bck = counts(srcImage, bck_ra, bck_dec, 
                                        bck_rad_in, bck_rad_out);

  /* Find ratio of source to background region areas and normalise background */
    double area_ratio = source.area / bck.area;
    double bck_counts = bck.counts * area_ratio;

  /* Find exposure at the source centre */
    double exposure = getExposure(exposurefile, src_ra,
                                    src_dec, srcImage.exposure);

  /* Find encircled energy factor for source region */
    double eef = realParameter("eef");  
    if (! booleanParameter("witheef"))   // calculate if not set as paremeter
       eef = getEEF(srcImage, src_rad);

  /* Calculate the error on the bckgnd subtracted count rate */
    double toterr = sqrt(source.counts + bck.counts
                                       * area_ratio*area_ratio) / eef;

  /* Calculate upper limit */
    double upplim = calcUpperlimit(exposure, source.counts,
                              bck_counts, toterr, ulsig, eef);

  /* Output useful debug if required */
    errstr<< "Source/back counts: "<<source.counts<<" "<<bck.counts<<endl
        << "Area ratio: "<<area_ratio<<endl
        << "Exposure: "<<exposure<<endl
        << "EEF: "<<eef
        << Sas::message(Msg::App, Msg::Verbose);

  /* Output the result */
    ostringstream outs;
    outs.precision(8);

    if (booleanParameter("details")) {
       outs<<"Source/back counts: "<<source.counts<<" "<<bck.counts<<endl
           <<"Area ratio: "<<area_ratio<<endl
           << "Exposure: "<<exposure<<endl
           << "EEF: "<<eef<<endl;
    }
    outs<<" Statistical upper limit c/r: "<<upplim<<" c/s" <<endl;

    if (booleanParameter("withoutputfile")) {

      /* Open a stream to an output text file */
       ofstream outfile;
       string textfile=stringParameter("output");
       outfile.open(textfile.c_str());
       outfile<<outs.str();
       outfile.close();
    
    } else { // Write to standard error

       errstr<<outs.str()
             <<Sas::message(Msg::App, Msg::Always);
    }
  // } REMOVE

  /* Set output parameters */

  ENDPROC;
}

/* Destructor */
Eupper:: ~Eupper()
{
  PROC;

    // CLOSE EVERYTHING DOWN PROPERLY
    //delete srcreg;            srcreg=0;

  ENDPROC;
};

/* Eupper::calcUpplim - calculates the upper limit for total counts
,n, background counts, b and confidence limit, cl */
double Eupper::calcUpplim(double n, double b, double cl)
{
   UpperLimitSolver u = UpperLimitSolver(static_cast<long int>(n),b);
   std::pair<double, double> limits = u.bayesian(cl);
   //double lower = limits.first, upper=limits.second;
   double upper=limits.second;

   //cout<<"LIMITS: "<<lower<<" "<<upper<<endl;

   RETURN upper;
}

Eupper::ImageData Eupper::readImage(const string & imagefile)
{
 /* Open the input image as a dataset */
    ImageData im;

    im.imageset = dataSetServer->open( imagefile, DataSet::Read );

   // Set up the WCS matrix - this is used internally within SSCLIB
   // conversion routines
    SscWcsConvertor::setWcs(imagefile);

  /* Check whether image axes are inverted */
 //   bool invert_axes = checkaxes(array);

  /* Make a pixel iterator from the input image */
  /* SHOULD handle inverted axes HERE !! */
    if (imagefile.find(":") != string::npos) {
       im.image = new ArrayPixelIterator(imagefile);
    } else {
       im.image = new ArrayPixelIterator(imagefile+":");
    }
    //cout<<"CLOCK: got the array"<<clock()<<endl;

  // Initialise the reading of this array
    im.image->first();

  /* Find the instrument and observing mode */
    im.instrument = im.imageset->attribute("INSTRUME")->asString();
    im.obsmode = im.imageset->attribute("OBS_MODE")->asString();

  /* Get the exposure from the main array */
    Array * array;
    if (imagefile.find(":") != string::npos) {
       array = im.imageset->array(imagefile);
    } else {
       array = im.imageset->array(imagefile+":");
    }

    if (array->hasAttribute("EXPOSURE")) {
       im.exposure = array->attribute("EXPOSURE")->asReal();
    } else {
       im.exposure = 1.0;
    }

   // If this is a detector image then we also need to set the 
   // spacraft and instrument attitude matrices
    if (im.imageset->hasAttribute("CTYPE1") ) {
       if (im.imageset->attribute("CTYPE1")->asString() == "DETX") {
          SscWcsConvertor::setAtt(imagefile); } }
    else if (array->hasAttribute("CTYPE1") ) {
       if (array->attribute("CTYPE1")->asString() == "DETX" ) {
          SscWcsConvertor::setAtt(imagefile); }}
 
  /* Return */
    return im;
}


/* Method - counts() - returns the total counts in the image region */
Eupper::COUNTS Eupper::counts(ImageData imdat, double ra, double dec,
                      double rad_inner, double rad_outer)
{
  PROC;

     Eupper::COUNTS result;

    // Find image pixel index which contains this RA, DEC
     PixelElement pixel = findPixel(imdat, ra, dec);

     int xcen = pixel.xIndex(), ycen = pixel.yIndex();

    // Calculate width of region in pixels
    // Ensure pixel width is in units of degrees
     double convtodeg=1.0;
     if (imdat.image->wcsconvertor()->xType().find("DET") != string::npos)
        convtodeg=72000.0;
 
     double xhwid = rad_outer / (abs(pixel.xWidth()) / convtodeg);
     double yhwid = rad_outer / (pixel.yWidth() / convtodeg);

     //cout<<"Pixel index: "<<xcen<<" "<<ycen<<" "<<xhwid
       //  <<" "<<yhwid<<endl;

     double rout_pix = xhwid, 
            rinn_pix = rad_inner / (abs(pixel.xWidth()) / convtodeg);

    // Find X and Y ranges of pixels
     int ixmin=pixel.xIndex() - INT(xhwid) - 1;
     int ixmax=pixel.xIndex() + INT(xhwid) + 1;
     int iymin=pixel.yIndex() - INT(yhwid) - 1;
     int iymax=pixel.yIndex() + INT(yhwid) + 1; 

     int xbins = imdat.image->xbins();
     int ybins = imdat.image->ybins();

     ixmin = min( max(1,ixmin), (xbins));
     iymin = min( max(1,iymin), (ybins));
     ixmax = min( max(1,ixmax), (xbins));
     iymax = min( max(1,iymax), (ybins));

     double total=0.0, offset=0.0;
     int npix_in_area=0;

     //cout<<"CLOCK:Iterating "<<clock()<<endl;
     //cout<<"Iterating over pixels: "<<ixmin<<" "<<ixmax<<" "<<iymin<<" "<<iymax<<endl;

   // Loop over the X,Y range which contains the region
     for (int iy = iymin; iy<=iymax; iy++)
     {
        for (int ix = ixmin; ix<=ixmax; ix++)
        {
        //   imdat.image->setIndex( (iy-1) * xbins + ix -1 );

          // Check if this pixel is within annulus
           offset = sqrt(float( ( xcen - ix )*( xcen - ix )
                   + ( ycen - iy )*( ycen - iy ) ));
           
           if (offset <= rout_pix && offset>= rinn_pix) {
              //total = total+pixel.value();
              unsigned long index = (iy-1) * xbins + ix -1;
              total = total+imdat.image->valueAt(index);
              npix_in_area++;
           }
           
           //cout<<"DEBUG: "<<pixel.value()<<" "<<iy<<endl;
        }
     }
     result.counts = total;
     result.area = double(npix_in_area);

     //cout<<"CLOCK:Stopped Iterating "<<clock();
     //cout<<"Counts are: "<<result.counts<<" "<<result.area<<endl;

  RETURN result;
}

double Eupper::getExposure(const string &exposurefile, double src_ra,
                              double src_dec, double defaultExp) 
{   
  PROC;

    double exposure = 0.0;
    if (exposurefile != "NotSet") {

      // Open the exposure file and get the data array
       ImageData expdata = readImage(exposurefile);

      // Get the pixel which contains the source centre
       PixelElement pixel = findPixel(expdata, src_ra, src_dec);

       exposure = pixel.value();

    } else {
       exposure = defaultExp;
    }

    return exposure;
}

double Eupper::calcUpperlimit(double exposure, double src_counts, 
                              double bck_counts, double toterr,
                              double cl, double eef) 
{
  // bck_counts is the background rate normalised to the source region area

   double upplim = 0.0;

   if (exposure > 0.0) {  // Leave upper limit as 0 if exposure <= 0
     if (src_counts < 80.0) { // Use Bayesian statistics
        upplim = calcUpplim(src_counts, bck_counts, cl);
     } else {// Use source counts + 2 x error

    /* Need to convert between confidence interval and sigma.
       Can't find the equation right now - so hardcode for 1,2 and 3-sigma */
       double fac=0.0;
       if ( cl>0.67 && cl<0.69)
          fac=1.0;
       else if ( cl>0.98)
          fac=3.0;
       else
          fac=2.0;

       upplim = ((src_counts - bck_counts) + fac * ( toterr * eef ));
     }
   }

  // Return the upper limit in c/s corrected for the encircled energy factor
   return upplim / eef / exposure;
}

double Eupper::getEEF(ImageData srcImage, double src_rad_deg) const
{
  // Hardwire these correction factors to allow the code to run on any image
  // and for any instrument
    double tol=1.0; // arcsecs
    double eef=1.0; // full coverage by default
    double src_rad = src_rad_deg * 3600.0; // In arcseconds
 
  // EPIC - values are for EPN
    if (srcImage.instrument.substr(0,3) == "EPN" || 
        srcImage.instrument.substr(0,4) == "EMOS" ) {

       if (srcImage.obsmode.substr(0,4) == "SLEW") {
          if ( abs(src_rad - 30.0) < tol ) { 
             eef=0.845;  // This is the slew PSF for 1.5 keV - not very E-dept.
          } else if ( abs(src_rad - 45.0) < tol ) {
             eef=0.907;
          } else if ( abs(src_rad - 60.0) < tol ) {
             eef=0.936;
          }
       } else { // POINTING we assume - set for 6' off-axis, otherwise slow.  
          if ( abs(src_rad - 30.0) < tol ) { 
             eef=0.876;  // This is the pointed PSF for 1.5 keV - not very E-dep
          } else if ( abs(src_rad - 45.0) < tol ) {
             eef=0.927;
          } else if ( abs(src_rad - 60.0) < tol ) {
             eef=0.951;
          }
       }
    } else if (srcImage.instrument.substr(0,4) == "PSPC") { // ROSAT
       if (srcImage.obsmode.substr(0,6) == "SURVEY") {
          if ( abs(src_rad - 30.0) < tol ) { 
             eef=0.5;  // This is TBD
          } else if ( abs(src_rad - 45.0) < tol ) {
             eef=0.5;
          } else if ( abs(src_rad - 60.0) < tol ) {
             eef=0.5;
          }
       } else { // Pointed data - VERY position-dependent
             errstr<<"No EEF value for ROSAT POINTED data"
                   << Sas::warning("DataNotSupported");
       }
    }

    return eef;
}

bool Eupper::checkaxes(Array *array) 
{
  PROC;

   /* If the X axis name is "Y" or "DETY" or "DEC" then the image
      is inverted */
    if ( array->attribute("CTYPE1")->asString().find("Y") != string::npos ||
         array->attribute("CTYPE1")->asString().find("DEC") != string::npos )
    {
       ALTRETURNV( true );
    }
    RETURN false;
}

// findPixel - a method to find the pixel which contains a given ra, dec
PixelElement Eupper::findPixel(Eupper::ImageData imdat, double ra, 
                                                   double dec) 
{
   // Input ra, dec should be in degrees - here we convert to radians
   double raR=ra*DEG_TO_RAD, decR=dec*DEG_TO_RAD;

   CoordPair p = SscWcsConvertor::raDecToImagePixels(raR, decR);

   int ix = floor(p.x()+0.5), iy=floor(p.y()+0.5);

   imdat.image->setIndex( (iy-1) * imdat.image->xbins() + ix -1 );
   PixelElement pixel = imdat.image->currentItem();

   return pixel;
}

