// Module: FindBestRmfName
// Author: Richard Saxton
//
// Description: 
//
// This class defines a section of an image and includes methods
// to manioulate it.
//
// Modifications: See ChangeLog
// 
// $Id: FindBestRmfName.cc,v 1.6 2011/02/03 16:15:47 rsaxton Exp $

// General classes
#include <cmath>

// SAS-specific classes
#include "errstr.h"
#include "Tracer.h"
#include "PixelIterator.h"
#include "ImagePixelIterator.h"
#include "TablePixelIterator.h"
#include "SourcePosition.h"
#include "DatafileAttributes.h"
#include "PosWCS.h"
#include "Xmm.h"
#include "DataFilter.h"
#include "DssBuilder.h"
#include "DataSubspace.h"
#include "Evaluator.h"
#include "metatask.h"
#include "attcalc.h"
#include "CalServer.h"
#include "CorePosData.h"

// General utility functions
#include "errstr.h"
#include "Tracer.h"
#include "NumericalRange.h"

// Local classes
#include "FindBestRmfName.h"


/* Constructor */
//FindBestRmfName::FindBestRmfName();
/* Destructor */
//FindBestRmfName::~FindBestRmfName()

/* Method: getName(sourceset) - finds the RMF corresponding to the spectrum */
string FindBestRmfName::getName( const string & sourceFile )
{
  PROC;

   string bestRmf = " ", patbit=" ", modebit=" ", posbit=" ";
   string intro=" ", datebit=" ";

 /* If any problems set a default string controlled by this bbolean */
   bool notfound = false;

 /* Open the input file as a dataset */
   DataSet *spectrumset = dataSetServer->open( sourceFile, DataSet::Read );
   Table * spectrumtable = spectrumset->table("SPECTRUM");

 /* Read header keywords */
   DatafileAttributes *fileAttribs = new DatafileAttributes(sourceFile);
   Xmm::Mode windowMode = fileAttribs->getWindowMode();

   
 /* Use rmfgen::CorePosData class to get the event grade and the 
    central source position */
   CorePosData * corePos = new CorePosData(spectrumset, spectrumtable,
                                                            fileAttribs);
   EventGrades eventGrades = corePos->patternCode();
           
   switch (fileAttribs->instrumentID())
    {
    case Xmm::EMOS1 :
    case Xmm::EMOS2 :
      {
         errstr << "Found event grade "<<eventGrades
                << Sas::message(Msg::App, Msg::Verbose);
           
       /* Instrument string */
         if (fileAttribs->instrumentID() == Xmm::EMOS1)
           { intro = "m1_";}
         else
           { intro = "m2_";}

       /* Get the pattern bits */
         switch (eventGrades)
          {
           case CombinedEventGrades::SINGLE :
            {
             patbit = "p0_"; break;
            }
           case CombinedEventGrades::SDTQ : 
            {
             patbit = "pall_"; break;
            }
           default:
            { notfound=true; }
          }

       /* Set the mode string */
         if ( windowMode == Xmm::FAST_UNCOMPRESSED ) // from somewhere
           { modebit = "tu"; }
         else
           { modebit = "im"; }
           
         errstr << "modebit "<<modebit
            << Sas::message(Msg::App, Msg::Verbose);
               
       /* Find whether this is in the centre, wings or off the patch. 
          At the moment the patch position is hard-coded. The alternative 
	  of getting it from the CCF is ambiguous */
         PixCoords cenRaw=corePos->centre_raw();
         int cenCCD=corePos->centre_CCD();

         posbit = 'o';   // Default is off patch

         if (cenCCD==1) {
            if ( (intro=="m1_" && //M1
                  (sqrt((cenRaw.x()-312)*(cenRaw.x()-312) +
		        (cenRaw.y()-293)*(cenRaw.y()-293) ) <=14 )) ||  

                 (intro=="m2_" && //M2
                  (sqrt((cenRaw.x()-301)*(cenRaw.x()-301) +
                        (cenRaw.y()-302)*(cenRaw.y()-302) ) <=14 )) 
               ) {
               posbit = 'c';  // centre of patch

            } else if ( (intro=="m1_" && //M1
                  (sqrt((cenRaw.x()-312)*(cenRaw.x()-312) +
                        (cenRaw.y()-293)*(cenRaw.y()-293) ) <=36 )) ||

                 (intro=="m2_" && //M2
                  (sqrt((cenRaw.x()-301)*(cenRaw.x()-301) +
                        (cenRaw.y()-302)*(cenRaw.y()-302) ) <=36 )) 
               ) {
               posbit = 'w';  // wings of patch
            }
         } // End of CCD=1 block

       /* Set the date string */
         STime numdate(fileAttribs->observationDate());   

         if ( numdate < STime("2000-10-03T00:00:00") )
           { datebit = "e1_"; }
         else if  ( numdate < STime("2001-04-22T00:00:00") )
           { datebit = "e2_"; }
         else if  ( numdate < STime("2001-11-07T00:00:00") )
           { datebit = "e3_"; }
         else if  ( numdate < STime("2002-05-26T00:00:00") )
           { datebit = "e4_"; }
         else if  ( numdate < STime("2002-11-05T00:00:00") )
           { datebit = "e5_"; }
         else if  ( numdate < STime("2004-01-14T00:00:00") )
           { datebit = "e6_"; }
         else if  ( numdate < STime("2005-02-14T00:00:00") )
           { datebit = "e7_"; }
         else if  ( numdate < STime("2006-03-22T00:00:00") )
           { datebit = "e8_"; }
         else if  ( numdate < STime("2007-04-24T00:00:00") )
           { datebit = "e9_"; }
         else if  ( numdate < STime("2008-05-28T00:00:00") )
           { datebit = "e10_"; }
         else if  ( numdate < STime("2009-07-01T00:00:00") )
           { datebit = "e11_"; }
         else if  ( numdate < STime("2010-08-03T00:00:00") )
           { datebit = "e12_"; }
         else 
           { datebit = "e13_"; }

                errstr << "DateBit "<<datebit
            << Sas::message(Msg::App, Msg::Verbose);
            
         //bestRmf = intro + modebit + "_" + patbit + "_" + datebit + ".rmf";
         //bestRmf = intro + datebit + modebit + "_" + patbit + "_" + "v1.2.rmf";
         bestRmf = intro + datebit + modebit + "_" + patbit + posbit + ".rmf";
         break;
      }
    case Xmm::EPN   :
      {
       /* Get the pattern bits */
         switch (eventGrades)
          {
           case CombinedEventGrades::SINGLE :
            {
             patbit = "s";
             break;
            }
           case CombinedEventGrades::SD :
            {
             patbit = "sd"; break;
            }
           case CombinedEventGrades::DOUBLE : 
            {
             patbit = "d"; break;
            }
           default:
            { notfound=true; }
          }

       /* Set the mode string */
         switch (windowMode) 
          {
           case Xmm::PRIME_FULL_WINDOW :
            { 
             modebit = "ff20"; break;
            }
           case Xmm::PRIME_FULL_WINDOW_EXTENDED :
            { 
             modebit = "ef20"; break;
            }
           case Xmm::PRIME_SMALL_WINDOW:
            { 
             modebit = "sw20"; break;
            }
           case Xmm::PRIME_LARGE_WINDOW:
            { 
             modebit = "lw20"; break;
            }
           case Xmm::FAST_TIMING :
            { 
             modebit = "ti40"; break;
            }
           case Xmm::FAST_BURST :
            { 
             modebit = "bu23"; break;
            }                                        
           default:
            { notfound=true; }
         }
         intro = "epn_";

       /* Find the RAW-Y position string */
         int pnline = INT(floor(corePos->centre_raw().y() / 20.0));
         posbit = "Y" + toString(pnline);

       /* Set the date string */
         STime numdate(fileAttribs->observationDate());

       // Dates are set to the epochs 2000-01-01, 2007-01-01, 2014-01-01
         if ( numdate < STime("2005-01-01T00:00:00") )
           { datebit = "e1_"; }
         else if  ( numdate < STime("2010-01-01T00:00:00") )
           { datebit = "e2_"; }
         else 
           { datebit = "e3_"; }
 
         //bestRmf = intro + modebit + "_" + patbit + posbit + "_v6.7.rmf";
         bestRmf = intro + datebit + modebit + "_" + patbit + posbit + ".rmf";
         break;
      }        
    default:
      errstr << "Dont recognise the instrument "<<fileAttribs->instrument()
             << Sas::error("invalidInstrument");
    } // end of instrument case

   /* Set default string if any problems during the name construction */
    if (notfound) {bestRmf = "NoCannedMatrixForThisSpectrum";}

    errstr << "Will set canned RMF name to "<<bestRmf
           << Sas::message(Msg::App, Msg::Verbose);
    
   /* Close down and tidy up */
    delete corePos;
    delete fileAttribs; 
    dataSetServer->close(spectrumset);
        
    RETURN bestRmf;
}

