// Module: ImageRegion
// Author: Richard Saxton
//
// Description: 
//
// This class defines a section of an image and includes methods
// to manioulate it.
//
// Modifications: See ChangeLog
// 
// $Id: ImageRegion.cc,v 1.22 2012/04/27 14:35:53 rsaxton Exp $

// General classes
#include <cmath>
//gcc4.8.2 portability
#include <unistd.h>

// SAS-specific classes
#include "errstr.h"
#include "Tracer.h"
#include "stringutil.h"
#include "PixelIterator.h"
#include "ImagePixelIterator.h"
#include "TablePixelIterator.h"
#include "SourcePosition.h"
#include "DatafileAttributes.h"
#include "PosWCS.h"
#include "Xmm.h"
#include "DataFilter.h"
#include "DssBuilder.h"
#include "DataSubspace.h"
#include "Evaluator.h"
//#include "metatask.h"
//#include "attcalc.h"
#include "CalServer.h"
#include "SscWcsConvertor.h"



// General utility functions
#include "errstr.h"
#include "Tracer.h"
#include "NumericalRange.h"

// Local classes
#include "ImageRegion.h"


/* Constructor */
ImageRegion::ImageRegion( const string & imagefile, const string & expression,
      const string & tempfilename, const Ccd ccdin):
  _filteredtablename(tempfilename),
  _imagefile(imagefile),
  _expr(expression),
  _init(false),
  _ccdin(ccdin),
  set(0)
{
  PROC;

  /* Set up filtered array etc. */
    init();

  ENDPROC;
};

/* Constructor without explicit CCD number */
ImageRegion::ImageRegion( const string & imagefile, const string & expression,
      const string & tempfilename):
		  _filteredtablename(tempfilename),
		  _imagefile(imagefile),
		  _expr(expression),
		  _init(false),
		  set(0)
{
  /* Initialise with CCD set to a dummy value */
	_ccdin=1;
	init();
};

/* Destructor */
ImageRegion::~ImageRegion()
{
  PROC;

  delete _image;      	        _image                = 0;
  //delete _fullimage;      	_fullimage            = 0;
  //delete _poswcs;               _poswcs               = 0;
  //delete _putils;               _putils               = 0;
  delete _regCentre;            _regCentre            = 0;
  // delete _array;            _array                    = 0;

 /* Close down opened files */
  //dataSetServer->close(_imageset);

  if (_init) {
     dataSetServer->close(_filteredset);
  }

  ENDPROC;

}

/* Method: init - for convenience put start-up code here */
void ImageRegion::init()
{
 PROC;

 /* Convert X,Y and DETX, DETY components of expression to upper case */
//      _expr = uppercase(_expr);
   upcaseunits(_expr);   

 /* Check that the input expression is valid and set variables*/
   interpretExpression(_expr);

 /* Open the input image as a dataset */
   _imageset = dataSetServer->open( _imagefile, DataSet::Read );

 /* Map the array extension */
    if (_imagefile.find(":") != string::npos) {
       _array = _imageset->array(_imagefile);
    } else {
       _array = _imageset->array(_imagefile+":");
    }

 /* Read header keywords */
    readkeywords();

  /* Set the image pixel coordinates and units */
    setUnits();

 /* Set the ssclib wcs and attitude structure */
    SscWcsConvertor::setWcs(_imagefile);
    SscWcsConvertor::setWcsRef(_imagefile);
    SscWcsConvertor::setAtt(_imagefile);

 /* Get the region centre as a sourcePosition object so that it can
    be accessed in various coordinate systems */
   //_putils = new PositionUtils(_imgAttribs->instrumentID());
   //_poswcs = new PosWCS(_imageset, _imgAttribs->rapnt(), _imgAttribs->decpnt(), _imgAttribs->papnt(), false);

  /* Convert input RAW to DET coords */
   if (_expCoordType == "RAW")
      convRawExprToDet();

   //cout<<"Expression now:"<<_expr<<" "<<_expCoordType<<endl;

  /* Check whether image axes are inverted */
    bool invert_axes = checkaxes();

  /* Make a pixel iterator from the input image */
    if (_imagefile.find(":") != string::npos) {
       _fullimage = new ImagePixelIterator(_imagefile, invert_axes);
    } else {
       _fullimage = new ImagePixelIterator(_imagefile+":", invert_axes);
    }

  /* Subset the input image using the expression to produce an array of
     pixels which are contained within the input region */
   _image = filterimage(_fullimage, _expr);

   errstr << "ImageRegion: constructed with "<<_expr
          << Sas::message(Msg::App, Msg::Verbose);

 /* Set internal extent and centre variables */
   setExtents();

   errstr<<" Setting up source pos: "<<_cenX<<" "<<_cenY<<" "<<_expCoordType
  	<<" "<<_imageCoordType<< Sas::message(Msg::App, Msg::Verbose);

//   _regCentre = new SourcePosition(_putils->
  //                         sourcePosition(_imageCoordType,
    //                                  _imgAttribs->datamode(),
      //                                _imgAttribs->observationDate(),
        //                              _cenX,
          //                            _cenY,
            //                          _poswcs));

  // Set the centre source position object which stores det and tel coords
   CoordPair detcen = conv2Det(centre());
   CoordPair tel = conv2Tel(centre());
   _regCentre = new SourcePosition(detcen.x(), detcen.y(), tel.x(), tel.y());

 ENDPROC;
}

/* Method: srcCentre - returns the centre of the source region in the coords
                       of the image */
CoordPair ImageRegion::centre() const
{
 PROC;

    RETURN CoordPair(_cenX, _cenY);
}

/* Method: centralCcd - returns the CCD containing the 
   centre of the source region in the units */
Ccd ImageRegion::centralCcd() const
{
 PROC;

    //GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);

    CoordPair pixel = conv2Det(centre());

    
    //CamCoords testpixel = _putils->detxy005ToCamCoord2(pixel.x(),
      //                                                      pixel.y());
   /* Find which CCD this is on */
    //Ccd ccd = ct->onCcd(testpixel.x(), testpixel.y());

    Ccd ccd = SscWcsConvertor::detToCcd(pixel.x(), pixel.y());
    // delete ct; ct=0;

    RETURN ccd;
}

/* Method: conv2Det - converts any input coords into DETXY */
CoordPair ImageRegion::conv2Det(CoordPair input) const
{
 PROC;

   // Convert input position to detector coords 
    if (_imageCoordType == "DET")
       RETURN input;
    else if (_imageCoordType == "POS") 
       RETURN SscWcsConvertor::xyToDet(input.x(), input.y());

   

 //   SourcePosition * sp = new SourcePosition(_putils->
   //                        sourcePosition(_imageCoordType,
     //                                 _imgAttribs->datamode(),
       //                               _imgAttribs->observationDate(),
         //                             input.x(),
           //                           input.y(),
             //                         _poswcs));

   // double x = sp->xdet(), y = sp->ydet();

   // delete sp; sp=0;

   // RETURN CoordPair(x,y);
}

/* Method: conv2Pos - converts any input coords (from the image) into XY */
CoordPair ImageRegion::conv2Pos(CoordPair input) const
{
 PROC;

    if (_imageCoordType == "POS")
       RETURN input;

    else if (_imageCoordType == "DET") 
       RETURN SscWcsConvertor::detToXy(input.x(), input.y());
}

/* Method: conv2Tel - converts any input coords (from image) into theta/phi */
CoordPair ImageRegion::conv2Tel(CoordPair input) const
{
 PROC;

    CoordPair detpos = conv2Det(input);
    RETURN SscWcsConvertor::detToTel(detpos.x(), detpos.y());
}

/* Method: conv2Raw - converts any input coords into RAW */
CoordPair ImageRegion::conv2Raw(CoordPair input) const
{
 PROC;

    CoordPair detpos = conv2Det(input);

    RETURN SscWcsConvertor::detToRaw(detpos.x(), detpos.y());

//    CamCoords camcoord2 = _putils->detxy005ToCamCoord2(
  //                                   INT(detpos.x()), INT(detpos.y()));

 //   PixCoords pixcoord = _putils->camCoord2ToPixCoord(camcoord2);

   // RETURN CoordPair(pixcoord.x(), pixcoord.y());
}

/* Method: calcNormal - calculates the angle subtended by a tangent at
azimuthal angle, phi, to the optical axis (degrees) */
double ImageRegion::calcNormal(double phi) const
{
 PROC;

  /* Calculate the orientation (degs) of the ellipse wrt the optical axis */
   double orient = (phi * 180.0 / M_PI) - 90.0;

  // If not in detector coords need to convert using the position angle
   if (_expCoordType != "DET")
      orient = _imgAttribs->papnt() - (90.0 + orient);

   //double pa = _imgAttribs->papnt();

   //cout<<"Calc: "<<((phi * 180.0 / M_PI) - 90.0)<<" "<<pa<<" "<<orient;

   // DOESNT WORK FOR XY !!!!!

   RETURN orient;
}

/* Method: getCoordType - finds the coord type from the select expression */
const string ImageRegion::getCoordType(const string & expr)
{
 PROC;

   string ctype=" ";

   string expression = uppercase(expr);

   if (expression.find("DET") != string::npos) {
      ctype = "DET";
   } else if (expression.find("RAW") != string::npos){
   	  ctype = "RAW";
   } else if (expression.find("DEC") != string::npos) {
      ctype = "EQPOS";
   } else {
      ctype = "POS";
   }

   RETURN ctype;
} 

/* Method: coordType - returns the coordinate type */
const string ImageRegion::coordType()
{
 PROC;

   RETURN _expCoordType;
}

/* Method: imcoordType - returns the coordinate type of the image */
const string ImageRegion::imcoordType()
{
 PROC;

   RETURN _imageCoordType;
}

/* Method: theta - returns the off-axis angle of this region */
double ImageRegion::theta() const
{
 PROC;

 RETURN _regCentre->theta();
}

/* Method: phi - returns the azimuthal angle of this region */
double ImageRegion::phi() const
{
 PROC;

 RETURN _regCentre->phi();
}

/* Method: celestialCentre - returns RA, DEC of source region centre*/
CoordPair ImageRegion::celestialCentre() const
{
 PROC;

   // Get centre in POS coords
    CoordPair posCentre = conv2Pos(centre());

   // Find RA,DEC at the centre
 //   CoordPair radec = _poswcs->wcs()->physicalCoords(posCentre.x(), 
   //                                                   posCentre.y());
    CoordPair radec = SscWcsConvertor::xyToRaDec(posCentre.x(), posCentre.y());
    double ra = radec.x() / DEG_TO_RAD;
    double dec = radec.y() / DEG_TO_RAD;
    if (ra > 360.0)
	   ra = ra - 360.0;
    else if (ra < 0.0)
       ra = ra + 360.0;

    RETURN CoordPair(ra,dec);
 
 ENDPROC;
}

/* Centroid this region */
CoordPair ImageRegion::centroid(const int Niters ) 
{
  PROC;

  // This is a bit duff because it only uses the pixels within the
  // source region rather than the whole image

     double xpix=0.0, ypix=0.0;
     double total=0.0, offset=0.0;
     int ngood=0;
    
   /* Set initial centre value */
     double xcen = _cenX;
     double ycen = _cenY;

   /* Find the radius in image units */
     double arad = getRadius() * asecToImgUnits();

     for (int i=1; i<Niters+1; i++)
     {

      /* Initialise */
        total=0.0;
        ngood=0;
        xpix=0.0, ypix=0.0;

      /* Loop over each region pixel */
        FOREACH(_image)
        {
          PixelElement pixel = _image->currentItem();

       /* Include this pixel if it lies within the extraction radius */
          offset = sqrt( ( xcen - pixel.xCentre() )*( xcen - pixel.xCentre() ) 
                   + ( ycen - pixel.yCentre() )*( ycen - pixel.yCentre() ) );

          if (offset <= arad)
          {
            total = total+pixel.value();
            xpix = xpix + (pixel.xCentre() - xWidth()/2.0) * pixel.value();
            ypix = ypix + (pixel.yCentre() - yWidth()/2.0) * pixel.value();
            ngood = ngood + 1;
          }

        }

        if (ngood > 0) {
           xpix = xpix / total;
           ypix = ypix / total;
        }
        xcen = xpix;
        ycen = ypix;
     }

     errstr 
  	<< "Centroided on "<<xcen<<" "<<ycen
  	<< Sas::message(Msg::App, Msg::Verbose);

  RETURN CoordPair(xcen, ycen);
}

/* Method - counts() - returns the total counts in the image region */
double ImageRegion::counts() const
{
  PROC; 

     double total=0;

     FOREACH(_image)
     {
        //PixelElement pixel = _image->currentItem();
        //total = total+pixel.value();
        total = total+_image->valueAt();
     }

  RETURN total;
}

/* Method - ccdList() - returns a list of the CCDs included in the region */
vector<Ccd> ImageRegion::ccdList() const
{
  PROC; 

     //GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);
     vector<Ccd> ccdlist(13);
     for (int i=0; i<=12; i++)
     {
        ccdlist[i]=0;
     }

     FOREACH(_image)
     {
        PixelElement inpixel = _image->currentItem();
        CoordPair pixel = conv2Det(CoordPair(inpixel.xCentre(),
                                             inpixel.yCentre()));
      /* Find which CCD this is on */
        Ccd ccd = SscWcsConvertor::detToCcd(pixel.x(), pixel.y());

      /* Set this CCD true in the list */
        ccdlist[ccd]=1;
     }

  RETURN ccdlist;
}

/* Method - centralPixelValue - returns the counts in the central pixel */
double ImageRegion::centralPixelValue() 
{
  PROC; 

     double value=0;
     //unsigned int i=0;

     FOREACH(_image)
     {
        //i++;

        PixelElement pixel = _image->currentItem();

        if (abs(pixel.xCentre() - _cenX) <= xWidth() &&
            abs(pixel.yCentre() - _cenY) <= yWidth() ) {
           value = pixel.value();
        }
        //if (i == _image->size() / 2 || i == (_image->size()+1)) {
          // PixelElement pixel = _image->currentItem();
           //value = pixel.value();
       // }
     }

  RETURN value;
}

/* Method flux() - returns the flux, mean counts per arcsec^2 */
double ImageRegion::flux()
{
  PROC; 

  RETURN (countsperpix() / pixelArea());
}

/* Method countsperpix() - returns the mean counts per pixel */
double ImageRegion::countsperpix()
{
  PROC; 

  RETURN (counts() / number_of_pixels());
}

/* Method number_of_pixels() - returns the number of pixels in the region */
double ImageRegion::number_of_pixels()
{
  PROC; 

  RETURN _image->size();
}

/* Method pixelArea() - returns the area of each image pixel in arcsecs */
double ImageRegion::pixelArea()
{
  PROC; 

  RETURN abs(xWidth() * yWidth()) * _xUnits * _yUnits;
}

/* Method area() - returns the area of the region in arcsec**2 */
double ImageRegion::area()
{
  PROC; 

  RETURN pixelArea() * number_of_pixels();
}

/* Method xWidth() - returns the xWidth of each image pixel in det/xy coords*/
double ImageRegion::xWidth()
{
  PROC; 

  RETURN _fullimage->wcsconvertor()->xWidth();
}

/* Method yWidth() - returns the yWidth of each image pixel in det/xy coords*/
double ImageRegion::yWidth()
{
  PROC; 

  RETURN _fullimage->wcsconvertor()->yWidth();
}

/* Method asecToImgUnits() - conversion factor between arcsec and image units */
double ImageRegion::asecToImgUnits()
{
  PROC; 

  RETURN (1.0 / _xUnits);
}

/* Method getRadius - returns the radius of the source region in arcsecs */
double ImageRegion::getRadius() const
{
  PROC; 

   /* Use the Y extents to calculate the radius - these dont suffer from
      the distortion of RA */

  RETURN (abs(_maxY - _minY) /2.0 * _yUnits);
}

/* Method exposure() - returns the exposure time for the image */
double ImageRegion::exposure() const
{
  PROC; 

  RETURN _exposure;
}

/* Method getRadProf - gets the radial profile from a position 
                       in the image out to a maximum radius */
ImageRegion::RadialProfile ImageRegion::getRadProf(double xcen, double ycen,
                                            double width, double maxrad) 
{
  PROC; 

      ImageRegion::RadialProfile r;

     /* Find the maximum radius to use in image units for a 5' radius */
      double maxoff = maxrad / _xUnits;
      double binwid = width / _xUnits;

     /* Set X axis values for radial profile */
      r.nbins = INT( maxoff / binwid ) + 1;  // Number of radial bins
      r.rwid = width;                    // Bin width (arcsec)

     /* Zero the data array */
      for (int i=0; i<r.nbins; i++) {
          r.data.push_back(0.0);
          r.error.push_back(0.0);
      }

//      cout<<"R: "<<r.nbins<<" "<<r.rwid<<" "<<binwid<<endl;

      vector<int> norm(r.nbins);

/* Loop over each element in the image (may need to subset first) */
      FOREACH(_fullimage)
      {
        PixelElement pixel = _fullimage->currentItem();

        double offset = sqrt( (pixel.xCentre()-xcen) * (pixel.xCentre()-xcen) 
                          + (pixel.yCentre()-ycen) * (pixel.yCentre()-ycen) );


      /* If pixel is within the required outer radius add to radprof bin */
      // Should also exclude if in chip gap etc...
        if (offset <= maxoff)
        {
           int bin = INT(offset / binwid);

           r.data[bin] += pixel.value();
           norm[bin]++;
        }
        //cout<<"C: "<<pixel.xCentre()<<" "<<xcen<<" "<<pixel.yCentre()<<" "<<
          //    ycen<<" "<<offset<<" "<<binwid<<endl; 
      }

    /* Normalise the output to per square arcsec */
      for (int i=0; i<r.nbins; i++) {
         if (norm[i] > 0) {
            r.error[i] = sqrt(r.data[i]) / (norm[i] * pixelArea());
            r.data[i] = r.data[i] / (norm[i] * pixelArea());
         }
//          cout<<"R3: "<<r.data[i]<<" "<<norm[i]<<" "<<pixelArea()<<endl;
      }

      
      RETURN r;
}


/* Method setExtents - find the X and Y limits of the spatial region */
const void ImageRegion::setExtents() 
{
  PROC; 

     _minX=100000000;
     _maxX=-100000000;
     _minY=100000000;
     _maxY=-100000000;

  /* Just find the max and min pixel positions in this array */
     FOREACH(_image)
     {
        PixelElement pixel = _image->currentItem();

        _minX = pixel.xCentre() < _minX ? pixel.xCentre():_minX;
        _maxX = pixel.xCentre() > _maxX ? pixel.xCentre():_maxX;
        _minY = pixel.yCentre() < _minY ? pixel.yCentre():_minY;
        _maxY = pixel.yCentre() > _maxY ? pixel.yCentre():_maxY;
     }

  /* Add in the extra half pixel */
     _minX = _minX - xWidth();
     _maxX = _maxX + xWidth();
     _minY = _minY - yWidth();
     _maxY = _maxY + yWidth();

     _cenX = (_minX + _maxX) / 2.0;
     _cenY = (_minY + _maxY) / 2.0;

     //cout<<"Extremes: "<<_minX<<" "<<_maxX<<endl;
  ENDPROC;
}
/* Method: filterimage - filters out pixels in the image which aren't
           contained within the selection expression */
TablePixelIterator * ImageRegion::filterimage(  ImagePixelIterator * fullimage, 
                                                   const string & expression )
{
  /* Create a container file to hold the temporary table */
    srand(getpid());
    bool filenotunique=true;
    string rawtablename    = "";
    while (filenotunique) {
       string postfix = toString(rand());
       rawtablename    = "rawpixellist" + postfix + ".ds:EVENTS";
       if (!dataSetServer->exists(rawtablename)) {
          set = dataSetServer->open(rawtablename, DataSet::Create);
          filenotunique=false;
       }
    }

  /* Subset the image array using the DSS extents */
  string Xaxis, Yaxis;
  double xmin, xmax, ymin, ymax;
  RangeBound lxb,uxb,lyb,uyb;

  if (_expCoordType == "DET") {
     Xaxis="DETX";
     Yaxis="DETY";
  }
  else if (_expCoordType == "EQPOS") {
     Xaxis="RA";
     Yaxis="DEC";
  } else {
     Xaxis="X";
     Yaxis="Y";
  }

 /* Set a DSS object */
  DataSubspace _dss = dssBuilder.getDss(expression);

  Range<double> extentX = _dss.realextent(Xaxis);
  Range<double> extentY = _dss.realextent(Yaxis);

  extentX.values(xmin, xmax, lxb,uxb);
  extentY.values(ymin, ymax, lyb,uyb);

  /* Define a subset of the image to use */
  int ixmin=0, ixmax=10000, iymin=0, iymax=10000;

  double xmid = (xmax+xmin) / 2.0, ymid= (ymax+ymin) / 2.0;

  /* Find the pixel containing the expression centre */
  CoordPair p, pmin, pmax; // Central image pixel

  if (_expCoordType == "EQPOS") {

     //double raR = xmid * DEG_TO_RAD;
     //double decR = ymid * DEG_TO_RAD;
     //p = SscWcsConvertor::raDecToImagePixels(raR, decR);
     //double xpixwid_deg = xWidth()/ 72000.;
     //double ypixwid_deg = yWidth()/ 72000.;
     //int src_xwid_pix = (int)ceil(abs(src_ra - xmin) / xpixwid_deg);  
     //int src_ywid_pix = (int)ceil(abs(src_dec - ymin) / ypixwid_deg);
     
     pmin = SscWcsConvertor::raDecToImagePixels(xmin * DEG_TO_RAD, 
                                                ymin * DEG_TO_RAD);
     pmax = SscWcsConvertor::raDecToImagePixels(xmax * DEG_TO_RAD, 
                                                ymax * DEG_TO_RAD);
 
  } else if (_expCoordType == "POS") {

     CoordPair cel = SscWcsConvertor::xyToRaDec(xmin, ymin);
     pmin = SscWcsConvertor::raDecToImagePixels(cel.x(), cel.y());
     cel = SscWcsConvertor::xyToRaDec(xmax, ymax);
     pmax = SscWcsConvertor::raDecToImagePixels(cel.x(), cel.y());

  } else if (_expCoordType == "DET") {

     CoordPair cel = SscWcsConvertor::detToRaDec(xmin, ymin);
     pmin = SscWcsConvertor::raDecToImagePixels(cel.x(), cel.y());
     cel = SscWcsConvertor::detToRaDec(xmax, ymax);
     pmax = SscWcsConvertor::raDecToImagePixels(cel.x(), cel.y());

  }
     
  // Convert to image pixel ranges
  ixmin = int(min(pmin.x(),pmax.x())) - 1;
  iymin = int(min(pmin.y(),pmax.y())) - 1;
  ixmax = int(max(pmin.x(),pmax.x())) + 1;
  iymax = int(max(pmin.y(),pmax.y())) + 1;

 /* Limit the range to be within the image */
  ixmin = ixmin < 1 ? 1 : ixmin;
  ixmin = ixmin > INT(fullimage->xbins()) ? INT(fullimage->xbins()) : ixmin;
  ixmax = ixmax < 1 ? 1 : ixmax;
  ixmax = ixmax > INT(fullimage->xbins()) ? INT(fullimage->xbins()) : ixmax;
  iymin = iymin < 1 ? 1 : iymin;
  iymin = iymin > INT(fullimage->ybins()) ? INT(fullimage->ybins()) : iymin;
  iymax = iymax < 1 ? 1 : iymax;
  iymax = iymax > INT(fullimage->ybins()) ? INT(fullimage->ybins()) : iymax;

  errstr<<"X : "<<xmin<<" "<<xmax<<" "<<ixmin<<" "<<ixmax
        <<"Y: "<<ymin<<" "<<ymax<<" "<<iymin<<" "<<iymax
        << Sas::message(Msg::App, Msg::Verbose);

  /* Set the expression to use */
  string tempexp = expression; // Variable to hold expression which may change

  /* If a simple circle or annulus convert the expression to X/Y */
  if (_expCoordType == "EQPOS") {
    CoordPair xycen = SscWcsConvertor::raDecToSkyPixels(xmid* DEG_TO_RAD, ymid* DEG_TO_RAD);
    tempexp = convExp(expression,xycen.x(),xycen.y());
  }

  unsigned subsize = (ixmax - ixmin +1) * (iymax - iymin +1);

  /* Add the table */
  Table * table = set->addTable(rawtablename, subsize);

  /* Add columns to this table */
    Column * valueCol  = table->addColumn("VALUE",
                                        Column::Real64, "Value", "");
 
    Column * skyxCol    = table->addColumn("X",
                                         Column::Int32, "Sky X", "");

    Column * skyyCol    = table->addColumn("Y", Column::Int32, "Sky Y", "");

    Column * eqxCol    = table->addColumn("RA",
                                         Column::Real64, "RA", "");

    Column * eqyCol    = table->addColumn("DEC", Column::Real64, "DEC", "");

    Column * detxCol    = table->addColumn("DETX", Column::Int16, 
                                             "x det coords", "0.05 arcsec");

    Column * detyCol    = table->addColumn("DETY", Column::Int16, 
                                             "y det coords", "0.05 arcsec");

    Column * indexxCol    = table->addColumn("XPIXEL", Column::Int16,
                                         "x index", "");

    Column * indexyCol    = table->addColumn("YPIXEL", Column::Int16,
                                         "y index", "");

  Column * timeCol    = table->addColumn("TIME",
                                         Column::Real64,
                                         "elasped time",
                                         "sec");

  ColumnData * timeColData   = timeCol->data();
  real64     * timeData      = timeColData->real64Data();

    ColumnData * valueColData  = valueCol->data();
    real64     * valueData     = valueColData->real64Data();

  ColumnData * skyxColData   = skyxCol->data();
  int32      * skyxData      = skyxColData->int32Data();

  ColumnData * skyyColData   = skyyCol->data();
  int32      * skyyData      = skyyColData->int32Data();

  ColumnData * eqxColData   = eqxCol->data();
  real64     * eqxData      = eqxColData->real64Data();

  ColumnData * eqyColData   = eqyCol->data();
  real64     * eqyData      = eqyColData->real64Data();

  ColumnData * detxColData   = detxCol->data();
  int16      * detxData      = detxColData->int16Data();

  ColumnData * detyColData   = detyCol->data();
  int16      * detyData      = detyColData->int16Data();

  ColumnData * indexxColData   = indexxCol->data();
  int16      * indexxData      = indexxColData->int16Data();

  ColumnData * indexyColData   = indexyCol->data();
  int16      * indexyData      = indexyColData->int16Data();

  /* Set the detector pixel size in degrees */
  double det2deg=1./(72000);

  /* Add WCS attributes to the DET columns (for attcalc) */
  WCSConvertor("DETX", 0, 0, det2deg, "DETY", 0, 0, det2deg).
                             addAttributesTo(detxCol, detyCol);

  /* Compute const time value to write to each row */
  double time = STime(_imageset->attribute("DATE-OBS")->asString())
                                            -Xmm::missionReferenceTime;

  /* Add WCS attributes to the index columns
     (for pixel-DET coord mapping; this way the WCS info
     can also store the pixel size) */
  fullimage->wcsconvertor()->addAttributesTo(indexxCol, indexyCol);

  /* Find the number of x bins */
  unsigned xbins = fullimage->xbins();

  /* Now iterate through input array adding a pixel entry for each */
  int row = 0;

  fullimage->first();

  CoordPair cel=0, sky=0, det=0;

  /* Only access the elements of interest */
  for (int iy = iymin; iy<=iymax; iy++)
  {
    fullimage->setIndex( (iy-1) * xbins + ixmin -1 );

    for (int ix = ixmin; ix<=ixmax; ix++)
    {
      PixelElement pixel = fullimage->currentItem();
      timeData[row] = time;
      valueData[row]  = pixel.value();

     // Make the conversions depending on the image type
      if (_imageCoordType == "POS") {

         cel = SscWcsConvertor::imagePixelsToRaDec(
                                                double(ix), double(iy));
         sky = SscWcsConvertor::raDecToSkyPixels(cel.x(), cel.y());
         det = SscWcsConvertor::raDecToDet(cel.x(), cel.y());

      } else { // DET coord image
         det = SscWcsConvertor::imagePixelsToDet(double(ix), 
                                                                double(iy));
         cel = SscWcsConvertor::detToRaDec(det.x(), det.y());
         sky = SscWcsConvertor::raDecToSkyPixels(cel.x(), cel.y());
      }
      skyxData[row] = floor(sky.x());
      //skyxData[row] = pixel.xCentre();
      skyyData[row] = floor(sky.y());
      detxData[row] = (int) det.x();
      detyData[row] = (int) det.y(); 
      eqxData[row]   = cel.x() / DEG_TO_RAD;
      eqyData[row]   = cel.y() / DEG_TO_RAD;

      indexxData[row] = pixel.xIndex();
      indexyData[row] = pixel.yIndex();
      ++row;

      //cout<<"Element: "<<ix<<" "<<iy<<" "
        //  <<"sky x: "<<sky.x()<<" "<<sky.y()
      //   <<"det: "<<det.x()<<" "<<det.y()
        //  <<"radec: "<<cel.x()<<" "<<cel.y()<<endl;

      errstr<<"added pixel: "<<pixel.xCentre()<<" "<<pixel.yCentre()
            <<"index["<<pixel.xIndex()<<","<<pixel.yIndex()<<"]"
                << Sas::message(Msg::App, Msg::Noisy);

     /* Move to next element if this isn't the last */
      if (! fullimage->isDone()) {
         fullimage->next();
      } 
    }
  }

  /* Close dataset  */
  delete indexyColData;
  delete indexxColData;
  delete detyColData;
  delete detxColData;
  delete skyyColData;
  delete skyxColData;
  delete valueColData;

  /* Run attcalc on this table if wanted */
  //if (_imageCoordType == "DET") {
    // runAttCalc(rawtablename);
  //}

  dataSetServer->close(set);

  /* Clone this so that we can work with it */
   _filteredset = dataSetServer->clone( rawtablename,
                                      _filteredtablename,
                                      Dal::Temp);

  /* Remove the raw table */
   dataSetServer->remove(rawtablename);

  /* Create a data filter from the input expression */
   //DataFilter filter(expression);
   DataFilter filter(tempexp);

   table = _filteredset->table("EVENTS");
   
   long rows = filter.doFilter(*table);

   errstr << "Found " << rows << " in filtered table."
          << Sas::message(Msg::App, Msg::Verbose);

  /* Set a flag to say that the filtered table has been created */
   _init=true;

  /* Create a new table filter object pointing to this table */
        RETURN (new TablePixelIterator( _filteredtablename,
                                "XPIXEL",
                                "YPIXEL",
                                "VALUE"
                                ));
}

/* Method: convExp - convert simple RA,DEC expression into X/Y */
string ImageRegion::convExp( const string & expression, 
                               double xpix, double ypix )
{
  PROC;

   string outexp = expression;

   vector<double> circleRad;
   double CFAC = 3600.*20.; // Conversion from degrees to X/Y pix

  // Convert expression to a DSS object
   DataSubspace dss = dssBuilder.getDss(expression);

  // Check whether there are one or more region filters 
   if (dss.components() == 1 && dss.component(0)->filters()==1) 
    {
    // Convert the one region filter into a GeoRegion object 
     GeoRegion region = dss.component(0)->filter(0)->region();
     
    // Check that expression only contains one filter
     if (region[0][0]->type() == ShapeBasics::CIRCLE)
         {
             double radius=region[0][0]->radii()[0]*CFAC;
             outexp = "(X,Y) in CIRCLE(" + toString(xpix) + "," 
                      + toString(ypix) + "," + toString(radius) + ")";
         } 
     else if (region[0][0]->type() == ShapeBasics::ANNULUS)
         {
             double inner = region[0][0]->radii()[0]*CFAC;
             double outer = region[0][0]->radii()[1]*CFAC;
             outexp = "(X,Y) in ANNULUS(" + toString(xpix) + "," 
                      + toString(ypix) + ","  
                      + toString(inner) + "," + toString(outer) + ")";
         }
    }
   //cout<<"Expression converted to "<<outexp<<endl;

    RETURN outexp;
      
  ENDPROC;
}

/* Method: interpretExpression - read and interpret the selection expression */
void ImageRegion::interpretExpression( const string & expression )
{
  PROC;

  /* Check that the input expression is valid */
    if (! Evaluator::isValidExpression(expression)) {
       errstr<<"Invalid expression: "<<expression
             << Sas::error("InvalidExpression");
    }

 /* Find and set the coordinate system from the expression */
   _expCoordType = getCoordType(expression);

 /* Set a DSS object */
//   DataSubspace _dss = dssBuilder.getDss(expression);

  ENDPROC;
}

/* Method: upcaseunits - attempt to fix problem of lower case X or DETX */
void ImageRegion::upcaseunits(string & expression)
{
  PROC;

   int pos = expression.find("x,y");
   if (pos != string::npos) 
      expression.replace(pos, 3, "X,Y");
   pos = expression.find("x, y");
   if (pos != string::npos) 
      expression.replace(pos, 4, "X, Y");
   pos = expression.find("detx,dety");
   if (pos != string::npos) 
      expression.replace(pos, 9, "DETX,DETY");
   pos = expression.find("detx, dety");
   if (pos != string::npos) 
      expression.replace(pos, 10, "DETX, DETY");
}

/* Method: setUnits - sets the units for the raw image pixels */
const void ImageRegion::setUnits()
{
  PROC;

   /* Attempt to get units from array extension */
   /* !!! Ignore this for the moment 
    if (_array->hasAttribute("CUNIT1"))
    {
       string sxunits = _array->attribute("CUNIT1")->asString();
       string syunits = _array->attribute("CUNIT2")->asString();

       if ( sxunits.find("deg") != string::npos) {
          _xUnits=3600.0;
          _yUnits=3600.0;
       }
    } else {
   */
     // Assume that these are detector or sky pixels
       _xUnits = 0.05;
       _yUnits = 0.05;
   // !!! }

    errstr<<" Units set to: "<<_xUnits<<" "<<_yUnits
          << Sas::message(Msg::App, Msg::Verbose);

   /* Set the image axes coordinates */
    if ( _array->attribute("CTYPE1")->asString().find("DET") != string::npos) 
    {
       _imageCoordType = "DET";
    } else {
       _imageCoordType = "POS";
    }

  ENDPROC;

}

const WCSConvertor * ImageRegion::setSkyWCS()
{
  PROC;

      double cdelt1 = _array->attribute("CDELT1")->asReal() /
                             _imageset->attribute("REFXCDLT")->asReal();
      double cdelt2 = _array->attribute("CDELT2")->asReal() /
                             _imageset->attribute("REFYCDLT")->asReal();

      const WCSConvertor * _wcs = new WCSConvertor( "Pixel",
                           (unsigned) _array->attribute("CRPIX1")->asInt() ,
                           _imageset->attribute("REFXCRPX")->asReal() ,
                           cdelt1,"Pixel",
                           (unsigned) _array->attribute("CRPIX2")->asInt() ,
                           _imageset->attribute("REFYCRPX")->asReal() ,
                           cdelt2);
  RETURN _wcs;
}

void ImageRegion::readkeywords()
{
  PROC;

   /* Read in the standard keywords */
    _imgAttribs = new DatafileAttributes(_imagefile);

   /* Get the exposure */
    if (_array->hasAttribute("EXPOSURE"))
    {
       _exposure = _array->attribute("EXPOSURE")->asReal();
    }
    else
    {
       errstr << "Exposure keyword not found - setting to 1000 seconds "
              << Sas::message(Msg::App, Msg::Sparse);
       _exposure = 1000.0;
    }

  ENDPROC;
}

const bool ImageRegion::checkaxes()
{
  PROC;

   /* If the X axis name is "Y" or "DETY" or "DEC" then the image
      is inverted */
    if ( _array->attribute("CTYPE1")->asString().find("Y") != string::npos ||
         _array->attribute("CTYPE1")->asString().find("DEC") != string::npos ) 
    {
       ALTRETURNV( true );
    }
    RETURN false;
}

// Method to return the height and width in sky/det pixels of the input image
CoordPair ImageRegion::fullImageExtent() 
{
    return CoordPair(_fullimage->xbins() * xWidth(), 
                     _fullimage->ybins() * yWidth());
}

// Method to get region extents
CoordPair ImageRegion::regionExtent()
{
    return CoordPair(abs(_maxX - _minX), abs(_maxY - _minY));
}

// Method to convert RAW input expression to DET
void ImageRegion::convRawExprToDet()
{

     // Extract raw centre from expression
     double rawx=0.0, rawy=0.0;  // TEST
	 unsigned index = _expr.find("CIRCLE("), indexc1 = _expr.find(",",index),
		   indexc2 = _expr.find(",",indexc1+1);

	 //cout<<"Index="<<index<<" "<<indexc1<<" "<<indexc2<<endl;
	 //cout<<"S1:"<<_expr.substr(index+7, (indexc1 - (index+7)))<<endl;
	 //cout<<"S2:"<<_expr.substr(indexc1+1, (indexc2 - (indexc1+1)))<<endl;

	 if (index != string::npos) {
        rawx = toNumber<double>(_expr.substr(index+7, (indexc1 - (index+7)) ));
	    rawy = toNumber<double>(_expr.substr(indexc1+1, (indexc2 - (indexc1+1)) ));
	 } else {
		errstr<<"Can not handle the expression: "<<_expr
		      <<Sas::error("InvalidRegion");
	 }
     // Set the CAL with the required CCD
     calServer.state()->ccd()->set(_ccdin);

     // Convert raw centre to detector coords (assumes CCD already set in CAL).
     //CoordPair detCoords = _putils->RawTodetxy005(rawx, rawy);
     CoordPair detCoords = SscWcsConvertor::rawToDet(rawx, rawy);
     
     // For now set a mickey-mouse radius
     string crad = "100.0";
     string srcexp = "(DETX,DETY) in CIRCLE(" + toString(detCoords.x()) + ","
                            + toString(detCoords.y()) + "," + crad + ")";

     // Reset the region expression
     _expr = srcexp;
     _expCoordType = "DET";
}

// findPixel - a method to find the pixel which contains a given ra, dec
PixelElement ImageRegion::findPixel(ImagePixelIterator *im, double ra,
                                                    double dec)
{
    // Input ra, dec should be in degrees - here we convert to radians
    double raR=ra*DEG_TO_RAD, decR=dec*DEG_TO_RAD;

    CoordPair p = SscWcsConvertor::raDecToImagePixels(raR, decR);

    int ix = floor(p.x()+0.5), iy=floor(p.y()+0.5);
    //cout<<"Pixels are: "<<p.x()<<" "<<p.y()<<" "<<ix<<" "<<iy<<endl;

    im->setIndex( (iy-1) * im->xbins() + ix -1 );
    PixelElement pixel = im->currentItem();

return pixel;
}

