// 
// Module: ImageRegion
// Author: Richard Saxton
//
// Description: 
//
// Given an input image, this class generates 
// a new PixelIterator for those pixels
// that are bounded by the specified region expression.
//
// Modifications: See ChangeLog
// 
// $Id: ImageRegion.h,v 1.12 2012/04/11 15:27:43 eojero Exp $

#ifndef ImageRegion_H
#define ImageRegion_H

// General classes
#include <iostream>

// SAS-specific classes
#include "PixelIterator.h"
#include "ImagePixelIterator.h"
#include "TablePixelIterator.h"
//#include "PositionUtils.h"
#include "SourcePosition.h"
#include "DatafileAttributes.h"
#include "Xmm.h"

// General utility functions

// Local classes

class ImageRegion
{
public:

  /* Constructor */
  ImageRegion( const string & imagefile, const string & expression,
                        const string & tempfilename, const Ccd ccdin);
  ImageRegion( const string & imagefile, const string & expression,
                         const string & tempfilename);

  /* Destructor */
  ~ImageRegion();

  /* Methods */
  //void runAttCalc(const string & rawtablename);
  double getRadius() const;
  double exposure() const;
  double counts() const;
  vector<Ccd> ccdList() const;
  double centralPixelValue();
  double countsperpix();
  CoordPair centroid(const int Niters );
  CoordPair conv2Det(CoordPair c) const;
  CoordPair conv2Raw(CoordPair c) const;
  CoordPair conv2Pos(CoordPair c) const;
  CoordPair conv2Tel(CoordPair c) const;
  CoordPair celestialCentre() const;
  CoordPair centre() const;
  CoordPair fullImageExtent();
  CoordPair regionExtent(); // Size of region in pixels
  Ccd centralCcd() const;
  const string getCoordType(const string & expression);
  const string coordType();
  const string imcoordType();
  double theta() const;
  double phi() const;
  double flux();
  double number_of_pixels();
  double pixelArea();
  double area();
  double xWidth();
  double yWidth();
  double asecToImgUnits();
  double calcNormal(double phi) const; 


  DatafileAttributes * attributes() {return _imgAttribs;}

  /* Structure */
  struct RadialProfile{
        vector<double> data;     // The radial profile
        vector<double> error;     // The radial profile
        double   rmin,           // Minimum radius
                 rwid;           // Width of each radial bin (arcsec)
        int      nbins;          // Number of radial bins
  };

  /* Get a radial profile from an image 
     xcen, ycen are the profile centre in the units of the image
     width is the width of each radial profile bin in arcseconds */
  RadialProfile getRadProf(double xcen, double ycen,
                                        double width, double maxrad);


  string convExp(const string & posexp,
                                     double xCen, double yCen);

  const WCSConvertor * setSkyWCS();

protected:

  const void setUnits();
  const bool checkaxes();

private:

  // Methods
  void convRawExprToDet();             // Converts raw expression to DET coords
  void init();
  void interpretExpression(const string & expression);
  void upcaseunits(string & expression);
  const void setExtents();
  void readkeywords();
  TablePixelIterator * filterimage(  
                   ImagePixelIterator * fullimage, const string & expression );

  PixelElement findPixel(ImagePixelIterator *im, double ra,
                                      double dec);
  DataSet * _filteredset;
  DatafileAttributes * _imgAttribs;
  string _expr;                        // The image region spatial expression
  string _filteredtablename; 
  string _imagefile;
  bool _init;
  DataSet * set;
  DataSet * _imageset;
  TablePixelIterator *  _image;
  double _minX;
  double _maxX;
  double _maxY;
  double _minY;
  double _cenX;
  double _cenY;
  ImagePixelIterator * _fullimage;
  //const PosWCS *      _poswcs;
  //PositionUtils * _putils;
  const SourcePosition * _regCentre;
  string _expCoordType;
  string _imageCoordType;
  Array * _array;
  double _xUnits;
  double _yUnits;
  string               _instrumentID;
  string               _telescope;
  string               _datamode;
  string               _dateobs;
  string               _dateend;
  Xmm::Instrument       _instr;
  int                  _emdhlow;
  int                  _emdhupp;

  double _rapnt;
  double _decpnt;
  double _papnt;
  double _exposure;

  Ccd _ccdin;

};

#endif
