//
// ecoordconv.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task to convert a point from any to all systems
//
//#include <ctime>
#include <unistd.h> // For getpid

#include "Tracer.h"
#include "errstr.h"
#include "param.h"

#include "CalServer.h"
#include "Dal.h"
#include "DssBuilder.h"

#include "ImageRegion.h"
#include "CalSetter.h"
#include "STime.h"
#include "SscWcsConvertor.h"
#include "ecoordconv.h"

void ecoordconv()
{
  PROC;

  /* Get the parameters from this command line */
    string imagefile = stringParameter("imageset");
    string srcexp = stringParameter("srcexp");
    bool pos2eqpos = booleanParameter("pos2eqpos");

  /* Attempt to set the CAL state using the input image header */
    CalSetter * _calset = new CalSetter();
    _calset->fromImage(imagefile);
    _calset->setAccuracyLevel(EXTENDED);

  /* If entered as separate position parameters read the params */
    string coordtype;
    double x=0, y=0;

    if  (booleanParameter("withcoords")) {
        x = realParameter("x");
        y = realParameter("y");
        coordtype = uppercase(stringParameter("coordtype"));

       // Allow a quick exit with pos->radec conversion only
        if (pos2eqpos && coordtype=="POS") {

           SscWcsConvertor::setWcsRef(imagefile);
           
           CoordPair c = SscWcsConvertor::xyToRaDec(x,y);

           errstr<<std::setprecision(8)<<" Region Centre: "<<endl
                 <<" RA: DEC: "<<c.x()/ DEG_TO_RAD<<" "<<c.y()/ DEG_TO_RAD 
                 << endl << Sas::message(Msg::App, Msg::Always);
           return; // leave now
        }

    } else {

     // Extract the coordinates from the src expression string
        srcexp = uppercase(stringParameter("srcexp"));

        if (srcexp.find("DET") != string::npos) {
           coordtype = "DET";
        } else if (srcexp.find("RAW") != string::npos){
           coordtype  = "RAW";
        } else if (srcexp.find("DEC") != string::npos) {
           coordtype = "EQPOS";
        } else {
           coordtype = "POS";
        }

        string Xaxis, Yaxis;
        double xmin, xmax, ymin, ymax;
        RangeBound lxb,uxb,lyb,uyb;

        if (coordtype == "DET") {
           Xaxis="DETX";
           Yaxis="DETY";
        } else if (coordtype == "EQPOS") {
           Xaxis="RA";
           Yaxis="DEC";
        } else if (coordtype == "RAW") {
           Xaxis="RAWX";
           Yaxis="RAWY";
        } else {
           Xaxis="X";
           Yaxis="Y";
        }

      /* Set a DSS object */
        DataSubspace _dss = dssBuilder.getDss(srcexp);

        Range<double> extentX = _dss.realextent(Xaxis);
        Range<double> extentY = _dss.realextent(Yaxis);

        extentX.values(xmin, xmax, lxb,uxb);
        extentY.values(ymin, ymax, lyb,uyb);

        x = (xmax+xmin) / 2.0; 
        y = (ymax+ymin) / 2.0;
    }

    // Set the CCD if RAW coordinates have been selected
    Ccd ccd=1;
    if (coordtype == "RAW") {
    	if (booleanParameter("withccd")) {
    	   ccd=intParameter("ccdno");
           calServer.state()->ccd()->set(ccd);     
    	} else {
    		errstr << "The CCD number (ccdno) must be set if "
    			   << "position is given in RAW coordinates"
    		             << Sas::error("noCCDNumber");
    	}
    }

  /* We can't currently optimise timing and burst mode data, so
     set status bad and exit early in these cases */
   /*
    Xmm::Mode windowMode = calServer.state()->mode()->get();

    if (windowMode == Xmm::FAST_TIMING ||
        windowMode == Xmm::FAST_BURST ||
        windowMode == Xmm::FAST_UNCOMPRESSED || 
        windowMode == Xmm::FAST_COMPRESSED) {
       
  //     setParameter("status",false);

       errstr<<" Timing/Burst mode data is not supported"
             << Sas::error("UnsupportedMode");

       ALTRETURN;
    }
  */

  /* Set the ssclib wcs and attitude structure */
    SscWcsConvertor::setWcs(imagefile);
    SscWcsConvertor::setWcsRef(imagefile);
    SscWcsConvertor::setAtt(imagefile);

  /* Initially convert to detector coordinates */
    CoordPair det(x,y), sky(x,y), raw(x,y), tel(x,y);
    CoordPair cel(x * DEG_TO_RAD,y * DEG_TO_RAD); // RA/DEC in radians
    bool onChip;

    if (coordtype == "DET") {
      // det = new CoordPair(x,y);
    } else if (coordtype == "POS") {
       //sky = new CoordPair(x,y);
       det = SscWcsConvertor::xyToDet(x,y);      
    } else if (coordtype == "EQPOS") {
       //cel = new CoordPair(x,y);
       det = SscWcsConvertor::raDecToDet(cel.x(), cel.y());
    } else if (uppercase(coordtype) == "RAW") {
       det = SscWcsConvertor::rawToDet(x,y);      
       onChip = true;
    } else {
            errstr << "Coordtype unknown: "<<coordtype
             << Sas::error("invalidCoordType");
    }

  /* Convert to celestial coords */
    if (uppercase(coordtype) != "EQPOS") 
       cel = SscWcsConvertor::detToRaDec(det.x(), det.y());

  /* Convert to Sky */
    if (uppercase(coordtype) != "POS") 
       sky = SscWcsConvertor::raDecToSkyPixels(cel.x(), cel.y());       
  
  /* Convert to Tel */
    tel = SscWcsConvertor::raDecToTel(cel.x(), cel.y());
    //tel = SscWcsConvertor::detToTel(det.x(), det.y());

  /* Convert to Raw */
    if (uppercase(coordtype) != "RAW") {
       raw = SscWcsConvertor::detToRaw(det.x(), det.y());

      /* Find the CCD for the source position */
       ccd = SscWcsConvertor::detToCcd(det.x(), det.y());
       //onChip = SscWcsConvertor::detOnChip(det.x(), det.y());
    }

  /* Find list of CCDs in source region */
    string ccds = "";

    if  (booleanParameter("withcoords")) {
        ccds = toString(ccd);
    } else { // Region - use an ImageRegion object
        srand(getpid());
        string tempf = "filtpixlist_src" + toString(rand()) + ".ds:";

        ImageRegion * srcreg = new ImageRegion(imagefile, srcexp, tempf, ccd);
        vector<Ccd> ccdlist = srcreg->ccdList();
        for (unsigned int i=1 ; i<ccdlist.size(); i++)
            if (ccdlist[i] != 0) ccds += " " + toString(i);

        delete srcreg; srcreg=0;
    }

  /* Set a few local variables */
    double theta  = tel.x();
    double phi = tel.y();

    double ra = cel.x() / DEG_TO_RAD;  // Converted to degrees
    double dec = cel.y() / DEG_TO_RAD;

    double detx = det.x();
    double dety = det.y();

    double rawx = raw.x();
    double rawy = raw.y();

    double posx = sky.x();
    double posy = sky.y();

  /* Output the result */
    errstr<<std::setprecision(8)<<" Region Centre: "<<endl
          <<" Theta: Phi: "<<theta<<" "<<phi<<endl
          <<" X: Y: "<<posx<<" "<<posy << endl
          <<" DETX: DETY: "<<detx<<" "<<dety << endl
          <<" RA: DEC: "<<ra<<" "<<dec << endl
          <<" RAWX: RAWY: "<<rawx<<" "<<rawy << endl
	  <<" CCD(s): "<<ccds<<" centred on CCD: "<<INT(ccd)<<endl
          << Sas::message(Msg::App, Msg::Always);

  /* Set output parameters */
    setParameter("theta_out",theta);
    setParameter("phi_out",phi);
    setParameter("posx_out",posx);
    setParameter("posy_out",posy);
    setParameter("detx_out",detx);
    setParameter("dety_out",dety);
    setParameter("rawx_out",rawx);
    setParameter("rawy_out",rawy);
    setParameter("ra_out",ra);
    setParameter("dec_out",dec);
    setParameter("ccd_out",INT(ccd));

  ENDPROC;
}

