//
// especget.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Metatask which runs evselect, arfgen and rmfgen to 
//              produce source and background spectra and the appropriate
//              ARF and RMF.
//
//#include <set>          // C++
//#include <algorithm>
//#include <strstream.h>
//#include <unistd.h>
#include <Tracer.h>
#include <metatask.h>
#include <errstr.h>
#include <param.h>
#include "Evaluator.h"

// SAS packages
#include "arfgen.h"   // tasks
#include "DataFilter.h"   // tasks
#include "DatafileAttributes.h"   // tasks

#include "especget.h"
#include "SpectrumMaker.h"
#include "ArfMaker.h"
#include "RmfMaker.h"
#include "BackscaleMaker.h"
#include "FindBestRmfName.h"
#include "ARF.h"
#include "EnergySpectrum.h"

void especget()
{
  PROC;

  /* Set the source, backgnd and arf filenames */
    string sourceset, backset, arfset, outrmfset;    

    if  (booleanParameter("withfilestem")) {
        string filestem = stringParameter("filestem");
        sourceset = filestem+"_src.ds";
        backset   = filestem+"_bgd.ds";
        arfset    = filestem+"_src.arf";
        outrmfset = filestem+"_src.rmf";
    } else {
        sourceset = stringParameter("srcspecset");
        backset   = stringParameter("bckspecset");
        arfset    = stringParameter("srcarfset");
        outrmfset = stringParameter("srcrmfset");
    }

  /* Get the parameters from this command line */
    string srcexp = stringParameter("srcexp");
    string backexp = stringParameter("backexp");
    bool withbadpixcorr = booleanParameter("withbadpixcorr");
    bool useodfatt = booleanParameter("useodfatt");
    string eventfile = stringParameter("table");
    bool withrmfset = booleanParameter("withrmfset");
    string rmfset = stringParameter("rmfset");
    bool extendedsource = booleanParameter("extendedsource");
    bool witharfset = booleanParameter("witharfset");
    bool withdetbins = booleanParameter("withdetbins");
    unsigned xbins = (unsigned) intParameter("detxbins");
    unsigned ybins = (unsigned) intParameter("detybins");
    bool multiEPICspec = booleanParameter("withmultiEPICspec");
    bool withenergybins = booleanParameter("withenergybins");
    float energymin = realParameter("energymin");
    float energymax = realParameter("energymax");
    int nenergybins = intParameter("nenergybins");
    bool withsourcepos = booleanParameter("withsourcepos");
    string sourcecoords = stringParameter("sourcecoords");
    float sourcex = realParameter("sourcex");
    float sourcey = realParameter("sourcey");
    bool a0removefiles = booleanParameter("a0removefiles");

  /* Read observation attributes from the event file */
    DatafileAttributes * specAttributes = new DatafileAttributes(eventfile);

  /* Check that the input expressions are valid */
    // First expand any macros 
    DataSet *set = dataSetServer->open(specAttributes->datafile(), Dal::Read);
    Table *table = set->table(0); 
    srcexp = DataFilter::expandSelectionExpression(srcexp,*table); 
    backexp = DataFilter::expandSelectionExpression(backexp,*table); 
    dataSetServer->close(set);

    if (! Evaluator::isValidExpression(srcexp)) {
       errstr<<"Invalid expression: "<<srcexp
             << Sas::error("InvalidExpression");
    }
    else if (! Evaluator::isValidExpression(backexp)) {
       errstr<<"Invalid expression: "<<backexp
             << Sas::error("InvalidExpression");
    }

  /* Create the source spectrum */
    errstr<<"Creating source spectrum..."
          << Sas::message(Msg::App, Msg::Sparse);

    SpectrumMaker sspec = SpectrumMaker(sourceset, specAttributes,
                                             srcexp,multiEPICspec);

  /* Create the background spectrum */
    errstr<<"Creating background spectrum..."
          << Sas::message(Msg::App, Msg::Sparse);

    SpectrumMaker bspec = SpectrumMaker(backset, specAttributes,
                                             backexp,multiEPICspec);

  /* Set the source area keyword (BACKSCAL) for the spectrum and
     create the ARF */
    bool createdRMF = false;
    if (witharfset) 
    {

      /* Set a boolean to control the use of an RMF within arfgen.
         It will be used unless a null rmfset name has been supplied. */
      bool useRmfWithinArfgen = true;

      /* Create the RMF if requested. Otherwise set to the given name*/
      if (! withrmfset) 
	{
	  errstr<<"Creating the RMF..."
		<< Sas::message(Msg::App, Msg::Sparse);
	  createdRMF = true;

	  RmfMaker(sourceset,outrmfset,extendedsource,
                       withdetbins, xbins, ybins,
                       withsourcepos, sourcecoords, sourcex, sourcey,
                       withenergybins, energymin,energymax,nenergybins);
	  
	  rmfset = outrmfset;
	}
      else 
	{  // If the rmfset name has not been given find the best default
	  if ( rmfset == " " ) {
	    rmfset = FindBestRmfName::getName(sourceset);
            useRmfWithinArfgen = false;  // Don't use an rmf within arfgen
	  }
	}

      errstr<<"Creating the ARF and calculating area for the source..."
	    << Sas::message(Msg::App, Msg::Sparse);

      ArfMaker(sourceset,
                       arfset,
                       withbadpixcorr,
                       useodfatt,
                       eventfile,
                       useRmfWithinArfgen,
                       rmfset,
                       extendedsource,
                       withdetbins, xbins, ybins,
                       withsourcepos, sourcecoords, sourcex, sourcey,
                       true);

      /* Set the background area keyword */
      errstr<<"Calculating area of the background region..."
          << Sas::message(Msg::App, Msg::Sparse);

      /* Make sure that the temporary file doesn't exist to avoid problems
	 with the dataset not being closed down correctly */
      string temparf =  "temparf.arf";
      if ( dataSetServer->exists(temparf))
	dataSetServer->remove(temparf);
      
      ArfMaker(backset,
                       temparf,
                       withbadpixcorr,
                       useodfatt,
                       eventfile,
                       useRmfWithinArfgen,
                       rmfset,
                       extendedsource,
                       withdetbins, xbins, ybins,
                       false, sourcecoords, sourcex, sourcey,
                       false);
    }

  /* Set the back, arf and rmf filenames in the source spectrum header */
    sspec.addkeywords(backset, rmfset, arfset);

  /* Set an output status variable */
    bool status = true;

  /* Check that the arf has been created with non-zero values */
    ARF *srcarf = new ARF(arfset);
    float arftot = srcarf->sum();
    errstr<<"Opened ARF: "<<srcarf->sum() << Sas::message(Msg::App, Msg::Verbose);
    delete srcarf; srcarf=0;

    if (arftot <= 0.0) {

      // Issue a warning
       errstr << "The effective areas contained in "<<arfset
              << " are zero or negative." 
                << Sas::warning("ZeroEffectiveArea");
     
      // Remove all files which have been produced
       if (a0removefiles) {
          dataSetServer->remove(sourceset);
          dataSetServer->remove(backset);
          if (witharfset)
             dataSetServer->remove(arfset);
          if (createdRMF)
             dataSetServer->remove(outrmfset);

          errstr <<"Removing all created spectra and matrices."
	         << Sas::message(Msg::App, Msg::Sparse);
       }

     /* Set the output status to say that the task has failed */
       status= false;
    }

    errstr <<"Setting the status variable" 
	  << Sas::message(Msg::App, Msg::Verbose);

   // Set the output parameter
    openParameters("especget");
    setParameter("status",status);

  /* Shutdown */

   //   closeParameters();

  ENDPROC;
}

